/*@ Internet Message Format (RFC 822 -> 2822 -> 5322) parser.
 *
 * Copyright (c) 2024 Steffen Nurpmeso <steffen@sdaoden.eu>.
 * SPDX-License-Identifier: ISC
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#ifndef su_IMF_H
#define su_IMF_H
#include <su/code.h>
#if defined su_HAVE_IMF || defined DOXYGEN /*XXX DOXYGEN bug; ifdef su_HAVE_IMF*/
#include <su/mem-bag.h>
#define su_HEADER
#include <su/code-in.h>
C_DECL_BEGIN
struct su_imf_addr;
struct su_imf_shtok;
#ifndef su_HAVE_MEM_BAG
# error IMF only with su_HAVE_MEM_BAG
#endif
/* su_imf* {{{ */
/* Whether we shall use a table lookup for char classification.
 * If value is S8_MAX we save array space with increased code size, with U8_MAX we simply lookup bytes */
#define su__IMF_TABLE_SIZE su_U8_MAX
enum su_imf_mode{
	su_IMF_MODE_NONE,
	su_IMF_MODE_OK_DISPLAY_NAME_DOT = 1u<<0,
	su_IMF_MODE_OK_ADDR_SPEC_NO_DOMAIN = 1u<<1,
	su_IMF_MODE_OK_DOT_ATEXT = 1u<<2,
	su_IMF_MODE_TOK_COMMENT = 1u<<3,
	su_IMF_MODE_TOK_SEMICOLON = 1u<<4,
	su_IMF_MODE_TOK_EMPTY = 1u<<5,
	su_IMF_MODE_RELAX = 1u<<6,
	su_IMF_MODE_STOP_EARLY = 1u<<7,
	su__IMF_MODE_ADDR_MASK = su_IMF_MODE_OK_DISPLAY_NAME_DOT | su_IMF_MODE_OK_ADDR_SPEC_NO_DOMAIN |
			su_IMF_MODE_RELAX | su_IMF_MODE_STOP_EARLY,
	su__IMF_MODE_STRUCT_MASK = su_IMF_MODE_OK_DOT_ATEXT | su_IMF_MODE_TOK_COMMENT |
			su_IMF_MODE_TOK_SEMICOLON | su_IMF_MODE_TOK_EMPTY |
			su_IMF_MODE_RELAX | su_IMF_MODE_STOP_EARLY,
	su__IMF_MODE_MASK = 0xFF
};
enum su_imf_state{
	su_IMF_STATE_DISPLAY_NAME_DOT = 1u<<8,
	su_IMF_STATE_ADDR_SPEC_NO_DOMAIN = 1u<<9,
	su_IMF_STATE_DOMAIN_LITERAL = 1u<<10,
	su_IMF_STATE_GROUP = 1u<<11,
	su_IMF_STATE_GROUP_START = 1u<<12,
	su_IMF_STATE_GROUP_END = 1u<<13,
	su_IMF_STATE_GROUP_EMPTY = 1u<<14,
	su_IMF_STATE_SEMICOLON = 1u<<15,
	su_IMF_STATE_COMMENT = 1u<<16,
	su_IMF_STATE_RELAX = 1u<<17,
	su_IMF_STATE_MASK = su_IMF_STATE_DISPLAY_NAME_DOT | su_IMF_STATE_ADDR_SPEC_NO_DOMAIN |
				su_IMF_STATE_DOMAIN_LITERAL | su_IMF_STATE_GROUP | su_IMF_STATE_GROUP_START |
				su_IMF_STATE_GROUP_END | su_IMF_STATE_GROUP_EMPTY |
			su_IMF_STATE_SEMICOLON | su_IMF_STATE_COMMENT
};
enum su_imf_err{
	su_IMF_ERR_GROUP_DISPLAY_NAME_EMPTY = 1u<<24,
	su_IMF_ERR_DISPLAY_NAME_DOT = 1u<<25,
	su_IMF_ERR_DQUOTE = 1u<<26,
	su_IMF_ERR_GROUP_OPEN = 1u<<27,
	su_IMF_ERR_COMMENT = 1u<<28,
	su_IMF_ERR_CONTENT = 1u<<29,
	su_IMF_ERR_RELAX = 1u<<30,
	su_IMF_ERR_MASK = su_IMF_ERR_GROUP_DISPLAY_NAME_EMPTY | su_IMF_ERR_DISPLAY_NAME_DOT |
				su_IMF_ERR_DQUOTE | su_IMF_ERR_GROUP_OPEN |
			su_IMF_ERR_COMMENT |
			su_IMF_ERR_CONTENT
};
/* Classification bits for su_imf_c_*() series */
enum su_imf_c_class{
	su_IMF_C_ALPHA = 1u<<0,
	su_IMF_C_DIGIT = 1u<<1,
	su_IMF_C_VCHAR = 1u<<2,
	su_IMF_C_ATEXT = 1u<<3,
	su_IMF_C_CTEXT = 1u<<4,
	su_IMF_C_DTEXT = 1u<<5,
	su_IMF_C_QTEXT = 1u<<6,
	su_IMF_C_SPECIAL = 1u<<7,
	su_IMF_C_NO_WS_CTL = 1u<<8,
	su_IMF_C_CR = 1u<<9,
	su_IMF_C_DQUOTE = 1u<<10,
	su_IMF_C_HT = 1u<<11,
	su_IMF_C_LF = 1u<<12,
	su_IMF_C_SP = 1u<<13
};
struct su_imf_addr{
	struct su_imf_addr *imfa_next;
	char *imfa_group_display_name;
	char *imfa_display_name;
	char *imfa_locpar;
	char *imfa_domain;
	char *imfa_comm;
	u32 imfa_group_display_name_len;
	u32 imfa_display_name_len;
	u32 imfa_locpar_len;
	u32 imfa_domain_len;
	u32 imfa_comm_len;
	u32 imfa_mse;
	char imfa_dat[VFIELD_SIZE(0)]; /* Storage for any text (single chunk struct) */
};
struct su_imf_shtok{
	struct su_imf_shtok *imfsht_next;
	u32 imfsht_mse;
	u32 imfsht_len;
	char imfsht_dat[VFIELD_SIZE(0)];
};
EXPORT_DATA u16 const su__imf_c_tbl[su__IMF_TABLE_SIZE + 1];
INLINE void *su_imf_snap_create(struct su_mem_bag *membp){
	ASSERT_RET(membp != NIL, NIL);
#ifdef su_HAVE_MEM_BAG_LOFI
	return su_mem_bag_lofi_snap_create(membp);
#else
	su_mem_bag_auto_snap_create(membp);
	return NIL;
#endif
}
INLINE void su_imf_snap_gut(struct su_mem_bag *membp, void *snap){
	ASSERT_RET_VOID(membp != NIL);
#ifdef su_HAVE_MEM_BAG_LOFI
	su_mem_bag_lofi_snap_gut(membp, snap);
#else
	UNUSED(snap);
	su_mem_bag_auto_snap_gut(membp);
#endif
}
EXPORT s32 su_imf_parse_addr_header(struct su_imf_addr **app, char const *header, BITENUM(u32,su_imf_mode) mode,
		struct su_mem_bag *membp, char const **endptr_or_nil);
EXPORT s32 su_imf_parse_struct_header(struct su_imf_shtok **shtpp, char const *header, BITENUM(u32,su_imf_mode) mode,
		struct su_mem_bag *membp, char const **endptr_or_nil);
#undef a_X
#if su__IMF_TABLE_SIZE == U8_MAX
# define a_X(X,Y) ((su__imf_c_tbl[S(u8,X)] & (Y)) != 0)
#elif su__IMF_TABLE_SIZE == S8_MAX
# define a_X(X,Y) (S(u8,X) <= S8_MAX && (su__imf_c_tbl[S(u8,X)] & (Y)) != 0)
#else
# error su__IMF_TABLE_SIZE must be U8_MAX or S8_MAX
#endif
SINLINE boole su_imf_c_ALPHA(char c) {return a_X(c, su_IMF_C_ALPHA);}
SINLINE boole su_imf_c_DIGIT(char c) {return a_X(c, su_IMF_C_DIGIT);}
SINLINE boole su_imf_c_VCHAR(char c) {return a_X(c, su_IMF_C_VCHAR);}
SINLINE boole su_imf_c_atext(char c) {return a_X(c, su_IMF_C_ATEXT);}
SINLINE boole su_imf_c_ctext(char c) {return a_X(c, su_IMF_C_CTEXT);}
SINLINE boole su_imf_c_dtext(char c) {return a_X(c, su_IMF_C_DTEXT);}
SINLINE boole su_imf_c_qtext(char c) {return a_X(c, su_IMF_C_QTEXT);}
SINLINE boole su_imf_c_special(char c) {return a_X(c, su_IMF_C_SPECIAL);}
SINLINE boole su_imf_c_obs_NO_WS_CTL(char c) {return a_X(c, su_IMF_C_NO_WS_CTL);}
SINLINE boole su_imf_c_CR(char c) {return a_X(c, su_IMF_C_CR);}
SINLINE boole su_imf_c_DQUOTE(char c) {return a_X(c, su_IMF_C_DQUOTE);}
SINLINE boole su_imf_c_HT(char c) {return a_X(c, su_IMF_C_HT);}
SINLINE boole su_imf_c_LF(char c) {return a_X(c, su_IMF_C_LF);}
SINLINE boole su_imf_c_SP(char c) {return a_X(c, su_IMF_C_SP);}
SINLINE boole su_imf_c_WSP(char c) {return (a_X(c, su_IMF_C_SP | su_IMF_C_HT));}
SINLINE boole su_imf_c_ANY_WSP(char c) {return (a_X(c, su_IMF_C_SP | su_IMF_C_HT | su_IMF_C_LF | su_IMF_C_CR));}
SINLINE boole su_imf_c_quoted_pair_c2(char c2){ /* NIL unsupported (manual) */
	return (a_X(c2, su_IMF_C_VCHAR | su_IMF_C_SP | su_IMF_C_HT | su_IMF_C_NO_WS_CTL | su_IMF_C_LF | su_IMF_C_CR));
}
#undef a_X
/* }}} */
C_DECL_END
#include <su/code-ou.h>
#endif /* su_HAVE_IMF */
#endif /* su_IMF_H */
/* s-itt-mode */
